/***********************************************************************
 * Tom Kelliher, Goucher College
 * Mar. 24, 2013
 * roomViewLab.cpp
 *
 * This is a demonstration program that draws a map of a simple room
 * and also draws a first-person view of the room, with some hooks
 * built-in to allow the user to navigate the room.
 ***********************************************************************/


const int ESC = 0x1b;


#include <math.h>
#include "Angel.h"


mat4 mv;   /* Model view matrix */


/* Identifiers for the shader programs and the uniform projection and model
 * view matrices in the vertex shader (see vshader41.glsl).
 */

GLuint program;
GLuint projection;
GLuint model_view;
GLuint vColor;
GLuint vPosition;


/* Basic data structures for drawing. */

typedef vec3 Color;
typedef vec3 Point;
typedef vec3 Position;


typedef struct Object
{
   GLuint vao;
   GLuint numVertices;
   GLuint geometry;
} Object;


Object oWall, oFloor, oViewer;


/* Initial window width and height */

GLuint windowWidth = 500;
GLuint windowHeight = 500;


// This wall section is 10 feet long (y) , 2 feet wide (x), and 8 feet
// high.  It is centered at the origin, in x and y and its base sits on
// the z = 0 plane.  All walls are constructed from this one section.

Point Wall[] =
{ Point(-1.0, -5.0, 0.0), Point( 1.0, -5.0, 0.0),
  Point( 1.0,  5.0, 0.0), Point(-1.0,  5.0, 0.0),
  Point(-1.0, -5.0, 8.0), Point( 1.0, -5.0, 8.0),
  Point( 1.0,  5.0, 8.0), Point(-1.0,  5.0, 8.0)
};


// Various colors.

Color red = Color(1.0, 0.0, 0.0);
Color green = Color(0.0, 1.0, 0.0);
Color blue = Color(0.0, 0.0, 1.0);
Color magenta = Color(1.0, 0.0, 1.0);
Color white = Color(1.0, 1.0, 1.0);
Color black = Color(0.0, 0.0, 0.0);


// This is a floor/ceiling section.  It's 24x24 and centered at the
//  origin.

Point Floor[] =
{ Point(-12.0, -12.0, 0.0), Point( 12.0, -12.0, 0.0),
  Point( 12.0,  12.0, 0.0), Point(-12.0,  12.0, 0.0)
};


// This is our roving viewer.  It's 1x1 and centered at the origin.

Point Viewer[] =
{ Point(-0.5, -0.5, 0.0), Point( 0.5, -0.5, 0.0),
  Point( 0.5,  0.5, 0.0), Point(-0.5,  0.5, 0.0)
};


Point points[36];   // Maximum points needed is 36.
GLuint Index;


// Constants necessary for rotating the viewer.  The trig functions take
// radian arguments.  (Sigh.)

const GLfloat ROTATE =   M_PI / 4.0; // 1/8th of a complete revolution in
                                     // radians.


// Viewer state.

GLfloat viewerTheta = M_PI / 2.0;    // View direction.  Initially looking 
                                     // up positive y axis.

Point viewerPosition = Point(0.0, 0.0, 6.0);   // Initial view position.


void quad( const Point vertices[], int a, int b, int c, int d );
void colorcube(const Point vertices[]);
GLuint createVAO(void);
void room();
void display(void);
void init(void);
void reshape(int w, int h);
void special(int key, int x, int y);
void keyboard(unsigned char key, int x, int y);


// quad generates two triangles for each face.

void quad(const Point vertices[], int a, int b, int c, int d )
{
    points[Index] = vertices[a]; Index++;
    points[Index] = vertices[b]; Index++;
    points[Index] = vertices[c]; Index++;
    points[Index] = vertices[a]; Index++;
    points[Index] = vertices[c]; Index++;
    points[Index] = vertices[d]; Index++;
}


// colorCube draws a cube.
//    vertices: the vertex list specifying the cube.
//
// Assumptions regarding the vertex list:
// Index   Vertex
//   0     Lower left vertex of bottom face.
//   1     Lower right vertex of bottom face.
//   2     Upper right vertex of bottom face.
//   3     Upper left vertex of bottom face.
// 4--7    Similar for top face.
// (Assumes we are looking at the origin from the +z axis with the +y axis
//  being "up."
//
// generate 12 triangles: 36 vertices.

void colorcube(const Point vertices[])
{
    quad( vertices, 1, 0, 3, 2 );
    quad( vertices, 2, 3, 7, 6 );
    quad( vertices, 3, 0, 4, 7 );
    quad( vertices, 6, 5, 1, 2 );
    quad( vertices, 4, 5, 6, 7 );
    quad( vertices, 5, 4, 0, 1 );
}


GLuint createVAO(void)
{
   GLuint vao, buffer;

   glGenVertexArrays(1, &vao);
   glBindVertexArray(vao);

   glGenBuffers(1, &buffer);
   glBindBuffer(GL_ARRAY_BUFFER, buffer);

   glBufferData(GL_ARRAY_BUFFER, sizeof(points),
                points, GL_STATIC_DRAW);

   glEnableVertexAttribArray(vPosition);
   glVertexAttribPointer(vPosition, 3, GL_FLOAT, GL_FALSE, 0,
                         BUFFER_OFFSET(0));

   return vao;
}


/***********************************************************************
 * room() --- draw the room.  Because of the separate needs of the map
 * and first-person views, it draws neither the ceiling nor the
 * viewer.
 ***********************************************************************/

void room()
{
   mat4 lmv;   // Local copy of the global modelview matrix.

   // Draw floor.
   glBindVertexArray(oFloor.vao);
   glUniform3fv(vColor, 1, white);
   glDrawArrays(oFloor.geometry, 0, oFloor.numVertices);

   // Draw middle walls.
   lmv = mv * Translate(-4.0, 0.0, 0.0);
   glUniformMatrix4fv(model_view, 1, GL_TRUE, lmv);
   glBindVertexArray(oWall.vao);
   glUniform3fv(vColor, 1, red);
   glDrawArrays(oWall.geometry, 0, oWall.numVertices);
   lmv = lmv * Translate(8.0, 0.0, 0.0);
   glUniformMatrix4fv(model_view, 1, GL_TRUE, lmv);
   glDrawArrays(oWall.geometry, 0, oWall.numVertices);

   // Draw walls defining room boundaries.

   lmv = mv * Translate(-11.0, -5.0, 0.0);
   glUniformMatrix4fv(model_view, 1, GL_TRUE, lmv);

   for (int i = 0; i < 4; i++)
   {
      glUniform3fv(vColor, 1, (i == 1) ? green : blue);
      glDrawArrays(oWall.geometry, 0, oWall.numVertices);
      lmv = lmv * Translate(0.0, 10.0, 0.0);
      glUniformMatrix4fv(model_view, 1, GL_TRUE, lmv);
      glUniform3fv(vColor, 1, (i == 1) ? green : blue);
      glDrawArrays(oWall.geometry, 0, oWall.numVertices);
      lmv = lmv * Translate(6.0, 6.0, 0.0) * RotateZ(-90.0);
      glUniformMatrix4fv(model_view, 1, GL_TRUE, lmv);
   }
}


/***********************************************************************
 * Recall, this will do our drawing for us.  It is called following
 * each simulation step in order to update the window.
 ***********************************************************************/

void display(void)
{
   mat4 p;    /* Projection matrix */

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   // Set-up to draw the map.

   p = Ortho(-13.0, 13.0, -13.0, 13.0, -10.0, 1.0);
   glUniformMatrix4fv(projection, 1, GL_TRUE, p);
      
   glViewport(0, 0, 100, 100);

   // Draw viewer.

   mv = Translate(viewerPosition[0], viewerPosition[1], 10.0);
   glUniformMatrix4fv(model_view, 1, GL_TRUE, mv);
   glBindVertexArray(oViewer.vao);
   glUniform3fv(vColor, 1, black);
   glDrawArrays(oViewer.geometry, 0, oViewer.numVertices);

   // Draw the room.

   mv = Translate(0.0, 0.0, 0.0);
   glUniformMatrix4fv(model_view, 1, GL_TRUE, mv);
   room();

   glFlush();

   // Set-up the do the first-person view.

   p = Perspective(90.0, 1.0, 0.1, 100.0);
   glUniformMatrix4fv(projection, 1, GL_TRUE, p);

   glViewport(100, 100, windowWidth - 100, windowHeight - 100);

   // Position the camera.
   mv = LookAt(vec4(viewerPosition[0], viewerPosition[1], viewerPosition[2],
                    1.0),
               vec4(viewerPosition[0] + cos(viewerTheta),
                    viewerPosition[1] + sin(viewerTheta),
                    viewerPosition[2], 1.0),
               vec4(0.0, 0.0, 1.0, 1.0));
   glUniformMatrix4fv(model_view, 1, GL_TRUE, mv);

    room();

   // Draw ceiling.
   mv = mv * Translate(0.0, 0.0, 8.0);
   glUniformMatrix4fv(model_view, 1, GL_TRUE, mv);
   glBindVertexArray(oFloor.vao);
   glUniform3fv(vColor, 1, magenta);
   glDrawArrays(oFloor.geometry, 0, oFloor.numVertices);

   glutSwapBuffers();
}


/***********************************************************************
 * init()
 ***********************************************************************/

void init(void) 
{
   glClearColor (0.0, 0.0, 0.0, 1.0);
   glShadeModel (GL_FLAT);
   glEnable(GL_DEPTH_TEST);

   /* Load shaders and use the resulting shader program. */

   program = InitShader("vshaderRVL.glsl", "fshaderRVL.glsl");
   glUseProgram(program);

   /* Get the locations of the uniform matrices in the vertex shader. */

   model_view = glGetUniformLocation( program, "model_view" );
   projection = glGetUniformLocation( program, "projection" );
   vColor = glGetUniformLocation( program, "vColor" );
   vPosition = glGetAttribLocation(program, "vPosition");

   Index = 0;
   colorcube(Wall);
   oWall.vao = createVAO();
   oWall.numVertices = 36;
   oWall.geometry = GL_TRIANGLES;

   Index = 0;
   quad(Floor, 0, 1, 2, 3);
   oFloor.vao = createVAO();
   oFloor.numVertices = 6;
   oFloor.geometry = GL_TRIANGLES;

   Index = 0;
   quad(Viewer, 0, 1, 2, 3);
   oViewer.vao = createVAO();
   oViewer.numVertices = 6;
   oViewer.geometry = GL_TRIANGLES;
}


/***********************************************************************
 * Callback for the non-ASCII keys.  I.e., the arrow keys.
 ***********************************************************************/

void special(int key, int x, int y)
{
   switch (key)
   {
   case GLUT_KEY_LEFT:
      --viewerPosition[0];
      break;

   case GLUT_KEY_RIGHT:
      ++viewerPosition[0];
      break;

   case GLUT_KEY_UP:
      ++viewerPosition[1];
      break;

   case GLUT_KEY_DOWN:
      --viewerPosition[1];
      break;

   default:
      return;
   }

   glutPostRedisplay();
}


/***********************************************************************
 * Hitting the Esc key will exit the program.
 *
 * "A" is rotate left and "S" is rotate right.
 ***********************************************************************/

void keyboard(unsigned char key, int x, int y)
{
   switch (key)
   {
   case ESC:
      exit(0);
      break;

   case 'a':
   case 'A':
      break;

   case 's':
   case 'S':
      break;

   default:
      return;
   }

   if (viewerTheta >= 2.0 * M_PI)
      viewerTheta -= 2.0 * M_PI;
   else if (viewerTheta < 0.0)
      viewerTheta += 2.0 * M_PI;

   glutPostRedisplay();

}


/***********************************************************************
 * Handler for reshape events.  This is done by either selecting the
 * largest viewport of the correct aspect ratio (in this case, 1:1) or
 * by reshaping the clipping rectangle so that its aspect ratio matches
 * that of the window.
 ***********************************************************************/

void reshape(int w, int h)
{
   windowWidth = w;
   windowHeight = h;

   glutPostRedisplay();
}


/***********************************************************************
 *  Request double buffer display mode for smooth animation.
 ***********************************************************************/

int main(int argc, char** argv)
{
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
   glutInitWindowSize(windowWidth, windowHeight); 
   glutInitWindowPosition(100, 100);
   glutInitContextVersion(3, 2);
   glutInitContextProfile(GLUT_CORE_PROFILE);
   glutCreateWindow("Room with a View");

   glewExperimental = GL_TRUE;
   glewInit();

   init();

   glutDisplayFunc(display); 
   glutReshapeFunc(reshape); 
   glutSpecialFunc(special);
   glutKeyboardFunc(keyboard);

   glutMainLoop();

   return 0;
}
